//! Handle NDScope style scope views
//! These views take an arbitrary number of channels and functions on those channels as a ViewSet
//! And also a ViewSpan, a time span, and continuously and asynchronously produce time domain results
//! across that time span for the ViewSet

mod pipeline_graph;

use std::collections::HashMap;
use std::hash::Hash;
use std::sync::atomic::{AtomicUsize, Ordering};
use gps_pip::{PipInstant, PipDuration};
use crate::analysis::graph::view_graph_to_pipeline::view_graph_to_pipeline;
use crate::analysis::result::ResultsReceiver;
use crate::data_source::{DataBlock, DataBlockReceiver, DataBlockSender, DataSourceRef};
use crate::errors::DTTError;
use crate::run_context::RunContext;
use crate::scope_view::pipeline_graph::create_pipeline_graph;
use tokio_util::{
    sync::{
        CancellationToken,
    },
};
#[cfg(not(feature = "python"))]
use dtt_macros::{staticmethod, getter, pyo3};

#[cfg(feature = "python")]
use pyo3::{
  pyclass, pymethods
};
use crate::analysis::scope::inline_fft::InlineFFTParams;
use crate::params::channel_params::channel::Channel;
use crate::user::ResponseToUser::ScopeViewResult;

#[derive(Clone, Debug)]
#[cfg_attr(feature = "python", pyclass)]
pub (crate) enum SetMember {
    Channel(Channel),
    /// The set has a name but no other channel information
    /// the channel name needs to be resolved into a Channel()
    /// before pipelines can be created. 
    UnresolvedChannelName(String),
    //Function(String, Vec<isize>),
}

#[derive(Clone, Debug)]
#[cfg_attr(feature = "python", pyclass)]
pub struct ViewSet {
    members: Vec<SetMember>,

    /// Index map by channel name
    /// for internal use only.
    hash: HashMap<String, usize>,
}

impl From<ViewSet> for Vec<Channel> {
    fn from(value: ViewSet) -> Self {
        let mut cvec = Vec::new();
        for m in value.members {
            if let SetMember::Channel(c) =  m {
                cvec.push(c)
            }
        }
        cvec
    }
}


impl From<Vec<Channel>> for ViewSet {
    fn from(value: Vec<Channel>) -> Self {
        Self {
            members: value.into_iter().map(|c| { SetMember::Channel(c) }).collect(),
            hash: HashMap::new()
        }
    }
}


impl From<Vec<String>> for ViewSet {
    fn from(value: Vec<String>) -> Self {
        Self {
            members: value.into_iter().map(|c| { SetMember::UnresolvedChannelName(c) }).collect(),
            hash: HashMap::new()
        }
    }
}

#[cfg_attr(feature = "python", pymethods)]
impl ViewSet {

    /// convenience function
    /// for turning a simple list of channels into a ViewSet
    #[staticmethod]
    pub fn from_channels(channels: Vec<Channel>) -> Self {
        channels.into()
    }

    /// convenience function
    /// for turning a simple list of channel names into a ViewSet with
    /// unresolved channel names
    #[staticmethod]
    pub fn from_channel_names(channel_names: Vec<String>) -> Self { channel_names.into() }
    
    
    pub fn has_unresolved_channels(&self) -> bool {
        self.members.iter().any(|m|
            matches!(m, SetMember::UnresolvedChannelName(_))
        )
    }

    /// Return the names of any channels in the set
    pub fn to_channel_names(&self) -> Vec<String> {
        let mut names = Vec::new();
        for member in self.members.iter() {
            match member {
                SetMember::UnresolvedChannelName(c) => names.push(c.clone()),
                SetMember::Channel(c) => names.push(c.name.clone()),
            };
        }
        names
    }
}

impl ViewSet {

    async fn update_hash(&mut self) {
        self.hash = self.members.iter().enumerate().map(|(i, m)| {
            (match m {
                SetMember::UnresolvedChannelName(n) => n.clone(),
                SetMember::Channel(c) => c.name.clone(),
            }
             , i
            )
        }).collect();
    }

    /// Change any unresolved channel names to resolved channels.
    async fn resolve_channels(&mut self, block: DataBlock) -> Option<DataBlock> {

        // hash the set



        println!("got a block in resolve channel");
        for channel in block.keys() {
            println!("looking for channel {}", channel.name);
            if let Some(i) = self.hash.get(channel.name.as_str()) {
                if matches!(self.members[*i], SetMember::UnresolvedChannelName(_)) {
                    println!("resolving channel {}", channel.name);
                    self.members[*i] = SetMember::Channel(channel.clone());
                }
            }
        }

        if self.has_unresolved_channels() {
            None
        } else {
            Some(block)
        }
    }
}

#[derive(Clone)]
pub (crate) enum SpanStart{
    Online,
    Offline(PipInstant),
}

/// When the span is online, the start is just a beginning marker taken
/// from the current time.
#[derive(Clone)]
pub (crate) struct ViewSpan {
    pub online: bool,
    pub start_pip: PipInstant,
    pub span_pip: PipDuration,
}

impl ViewSpan {
    pub (crate) fn optional_end_pip(&self) -> Option<PipInstant> {
        if self.online {
            None
        } else {
            Some(self.end_pip())
        }
    }

    pub (crate) fn end_pip(&self) -> PipInstant {
        self.start_pip + self.span_pip
    }


    fn new_fixed(start_pip: PipInstant, end_pip: PipInstant) -> Self {
        Self {
            online: false,
            start_pip,
            span_pip: end_pip - start_pip,
        }
    }

    fn new_online(span: PipDuration) -> ViewSpan {

        let start_pip = PipInstant::from_gpst_seconds(0.0);

        ViewSpan {
            online: true,
            start_pip,
            span_pip: span,
        }

    }
}


enum GapHandling {
    NaN,
    Fill(f64),
    Max,
    Min,
    Zero,
    None,
}

enum ScopeViewCommand {
    SetFFTParams{params: InlineFFTParams},
    SetSpan{span: ViewSpan},
}

pub struct ScopeView {
    pub (crate) id: usize,
    pub (crate) set: ViewSet,
    pub (crate) span: ViewSpan,
    pub (crate) gap_handling: Vec<GapHandling>,

    /// Once created, this value can be used to
    /// Update the scope window
    pub (crate) block_tx: Option<DataBlockSender>,
    
    /// Used to cancel the data task if the view is updated
    pub (crate) data_task_cancel_token: Option<CancellationToken>,
    
    /// Used to update a span of unordered live request in place
    pub (crate) span_update_tx: Option<tokio::sync::watch::Sender<PipDuration>>,

    /// Used to update frequency domain configuration for on-the-fly
    /// frequency domain results
    pub (crate) fft_config_tx: tokio::sync::watch::Sender<InlineFFTParams>,

    /// Number of times span has been hard-reset (as opposed to just updated)
    /// useful for debugging.
    pub (crate) reset_count: u64,
    
    /// used to change the view while running
    pub (crate) command_rx: tokio::sync::mpsc::UnboundedReceiver<ScopeViewCommand>,
}

impl ScopeView {
    pub (crate) fn create(rc: Box<RunContext>, data_source: DataSourceRef, id: usize, set: ViewSet, span: ViewSpan) -> tokio::sync::mpsc::UnboundedSender<ScopeViewCommand> {
        let (fft_config_tx, _) = tokio::sync::watch::channel(InlineFFTParams::default());
        let (command_tx, command_rx) = tokio::sync::mpsc::unbounded_channel();
        let view = Self {
            id,
            set,
            span,
            block_tx: None,
            data_task_cancel_token: None,
            span_update_tx: None,
            fft_config_tx,
            gap_handling: Vec::new(),
            reset_count: 0,
            command_rx,
        };

        tokio::spawn(
            view.run_loop(rc, data_source)
        );

        command_tx
    }

    async fn setup_analysis(& mut self, rc: &Box<RunContext>,
                                        block_rx: DataBlockReceiver)
                                        -> Result<ResultsReceiver, DTTError>
    {

        let mut ag = create_pipeline_graph(rc.clone(), self)?;
        
        view_graph_to_pipeline(rc, self, &mut ag,  block_rx).await
    }

    async fn start_results_loop(&mut self, rc: &Box<RunContext>, rr: ResultsReceiver) {

        let ct = CancellationToken::new();
        let ct2 = ct.clone();
        tokio::spawn( ScopeView::results_loop(rc.clone(), self.id, rr, ct2) );
    }

    async fn results_loop(rc: Box<RunContext>, id: usize, mut rr: ResultsReceiver, cancel_token: CancellationToken)  {
        loop {
            tokio::select! {
                _ = cancel_token.cancelled() => {
                    break
                },
                r = rr.recv() => {
                    match r {
                        Some(m) => if rc.output_handle.send(rc.clone(), ScopeViewResult{id, result: m}).is_err() {
                                   break;
                        },
                        None => {
                            break;
                        }
                    }
                },
            }
        }
    }

    fn set_span(&mut self, rc: &Box<RunContext>, data_source: DataSourceRef, span: ViewSpan) -> Result<(), DTTError> {
        if span.online == self.span.online {
            self.span = span;
            data_source.clone().update_scope_data(rc.clone(), self)
        } else {
            Err(DTTError::ViewConfig("Cannot switch an existing view between online and fixed.  Create a new view instead.".into()))
        }
    }
    
    fn set_fft_params(&mut self, params: InlineFFTParams) {
        let _ = self.fft_config_tx.send(params);
    }

    /// If the return value  is true, close the view
    async fn handle_command(&mut self, rc: &Box<RunContext>, data_source: DataSourceRef, command: ScopeViewCommand) -> bool {
        match command {
            ScopeViewCommand::SetFFTParams{params} => {
                self.set_fft_params(params);
            },
            ScopeViewCommand::SetSpan {span} => {
                if let Err(e) = self.set_span(rc, data_source, span) {
                    rc.user_messages.error(format!("Error setting span: {}", e));
                }
            }
        }
        false
    }
    
    async fn run_loop(mut self, rc: Box<RunContext>, data_source: DataSourceRef) {

        println!("View {} Started", self.id);
        // get run context

        println!("View {} got a run context", self.id);

        let mut block_rx = match data_source.clone().start_scope_data(rc.clone(), &mut self) {
            Ok(b) => b,
            Err(e) => {
                rc.clone().user_messages.error(format!("Error starting data source: {}", e));
                return;
            }
        };

        let mut fuse_block_channels = false;
        let (block_splice_tx, block_splice_rx) = tokio::sync::mpsc::channel(1);

        let mut bs_rx = Some(block_splice_rx);

        self.set.update_hash().await;

        loop {
            tokio::select! {
                r = block_rx.recv() => {
                    match r {
                        Some(b) => {
                            if fuse_block_channels {
                                if block_splice_tx.send(b).await.is_err() {
                                    break;
                                };
                            } else {
                                match self.set.resolve_channels(b).await {
                                    Some(b) => {
                                        println!("View {} Resolved Channels", self.id);
                                        fuse_block_channels = true;
                                        let rx = match bs_rx {
                                            Some(rx) => rx,
                                            None => {
                                                rc.clone().user_messages.error("Error resolving channels: block receive channel already used");
                                                break;
                                            }
                                        };
                                        let rr = match self.setup_analysis(&rc, rx).await {
                                          Ok(rr) => rr,
                                          Err(e) => {
                                                rc.clone().user_messages.error(format!("View {}: Error setting up analysis: {}", self.id, e));
                                                break;
                                            }
                                        };
                                        self.start_results_loop(&rc, rr).await;
                                        bs_rx = None;
                                        if block_splice_tx.send(b).await.is_err() {
                                            break;
                                        };
                                    },
                                    None => {
                                        println!("View {}: not all channels resolved", self.id);
                                    },
                                }
                            }
                        },
                        None => {break;}
                    }
                }
                r = self.command_rx.recv() => {
                    match r {
                        None => break,
                        Some(c) => {
                            if self.handle_command(&rc, data_source.clone(), c).await {
                                break;
                            }
                        }
                    }
                }
            }
        }
        println!("View {} Closed", self.id);
    }
}


static NEXT_SCOPE_VIEW_ID: AtomicUsize = AtomicUsize::new(0);


/// Clonable handle used to update a view
#[cfg_attr(feature = "python", pyclass(frozen))]
#[derive(Clone, Debug)]
pub struct ScopeViewHandle {
    id: usize,
    command_tx: tokio::sync::mpsc::UnboundedSender<ScopeViewCommand>,
}

impl ScopeViewHandle {

    fn set_span(&self, span: ViewSpan) -> Result<(), DTTError> {
        self.command_tx.send(ScopeViewCommand::SetSpan { span }).map_err(|x| x.into())
    }

    pub (crate) fn new_online(rc: Box::<RunContext>, data_source: DataSourceRef, set: ViewSet, span_pip: PipDuration) -> Self {
        let span = ViewSpan::new_online(span_pip);
        let id = NEXT_SCOPE_VIEW_ID.fetch_add(1, Ordering::Relaxed);
        let command_tx = ScopeView::create(rc, data_source, id, set, span);
        Self {
            id,
            command_tx,
        }
    }

    pub (crate) fn new_fixed(rc: Box::<RunContext>, data_source: DataSourceRef, set: ViewSet, start_pip: PipInstant, end_pip: PipInstant) -> Self {
        let span = ViewSpan::new_fixed(start_pip, end_pip);
        let id = NEXT_SCOPE_VIEW_ID.fetch_add(1, Ordering::Relaxed);
        let command_tx = ScopeView::create(rc, data_source, id, set, span);
        Self {
            id,
            command_tx,
        }
    }

    pub fn update_online(&self, span_pip: PipDuration) -> Result<(), DTTError> {
        let span = ViewSpan::new_online(span_pip);
        self.set_span(span)
    }

    pub fn update_fixed(&self, start_pip: PipInstant, end_pip: PipInstant) -> Result<(), DTTError> {
        let span = ViewSpan::new_fixed(start_pip, end_pip);
        self.set_span(span)
    }
}

#[cfg_attr(feature = "python", pymethods)]
impl ScopeViewHandle {
    #[getter]
    pub fn id(&self) -> usize {
        self.id
    }

    pub fn set_fft_params(& self, params: InlineFFTParams) -> Result<(), DTTError>{
        self.command_tx.send(ScopeViewCommand::SetFFTParams{params}).map_err(|x| x.into())
    }


    /// Update the span of a view.  This is a pythonic wrapper.  Rust code should call update_online() or update_fixed().
    //#[cfg(python)]
    //#[cfg_attr(feature="python", pyo3(text_signature = "(data_source, span_pip=None, start_pip=None, end_pip=None)"))]
    //#[pyo3(text_signature = "(span_pip: PipDuration=None, start_pip: PipInstant=None, end_pip: PipInstant=None)")]
    #[pyo3(signature = (span_pip=None, start_pip=None, end_pip=None))]
    pub fn update(&self, span_pip: Option<PipDuration>, start_pip: Option<PipInstant>, end_pip: Option<PipInstant>) -> Result<(), DTTError> {
        match (span_pip, start_pip, end_pip) {
            (Some(span_pip), Some(start_pip), None) => {
                self.update_fixed(start_pip, start_pip + span_pip)
            },
            (None, Some(start_pip), Some(end_pip)) => {
                self.update_fixed(start_pip, end_pip)
            },
            (Some(span_pip), None, None) => {
                self.update_online(span_pip)
            },
            _ => {
                Err(DTTError::ViewConfig("Invalid view configuration.  For an online span, set only span_pip.  For a fixed span, set start_pip and either span_pip or end_pip.".into()))
            },
        }
    }
}